(*
 * This file defines our counter, which we use in the Count command.
 *)

(*
 * Increment the counter
 *)
val increment : unit -> unit

(*
 * Determine the value of the counter
 *)
val value : unit -> int
